class Manga:
    """Represents a MangaDex Manga."""
    __slots__ = ("id", "title", "titles", "desc", "locked", "links", "language", "last_volume", "last_chapter",
                 "type", "status", "year", "content", "tags", "created_at", "updated_at", "author", "artist",
                 "chapters", "client")

    def __init__(self, data, rel, client):
        self.id = data["id"]
        self.title = data["attributes"]["title"]
        self.titles = data["attributes"]["altTitles"]
        self.desc = data["attributes"]["description"]
        self.locked = data["attributes"]["isLocked"]
        self.links = data["attributes"]["links"]
        self.language = data["attributes"]["originalLanguage"]
        self.last_volume = data["attributes"]["lastVolume"]
        self.last_chapter = data["attributes"]["lastChapter"]
        self.type = data["attributes"]["publicationDemographic"]
        self.status = data["attributes"]["status"]
        self.year = data["attributes"]["year"]
        self.content = data["attributes"]["contentRating"]
        self.tags = [MangaTag(x) for x in data["attributes"]["tags"]]
        self.created_at = data["attributes"]["createdAt"]
        self.updated_at = data["attributes"]["updatedAt"]
        self.author = [x["id"] for x in rel if x["type"] == "author"]
        self.artist = [x["id"] for x in rel if x["type"] == "artist"]
        self.client = client

    def get_chapters(self, params=None):
        return self.client.get_manga_chapters(self, params)


class MangaTag:
    """Represents a MangaDex Manga Tag."""
    __slots__ = ("id", "name")

    def __init__(self, data):
        self.id = data["id"]
        self.name = data["attributes"]["name"]



class Chapter:
    """Represents a MangaDex Chapter."""
    __slots__ = ("id", "volume", "chapter", "title", "language", "hash", "pages", "pages_redux", "published_at",
                 "created_at", "updated_at", "parent_manga", "group", "uploader", "client")

    def __init__(self, data, rel, client):
        self.id = data["id"]
        self.volume = data["attributes"]["volume"]
        self.chapter = data["attributes"]["chapter"]
        self.title = data["attributes"]["title"]
        self.language = data["attributes"]["translatedLanguage"]
        self.hash = data["attributes"]["hash"]
        self.pages = data["attributes"]["data"]
        self.pages_redux = data["attributes"]["dataSaver"]
        self.published_at = data["attributes"]["publishAt"]
        self.created_at = data["attributes"]["createdAt"]
        self.updated_at = data["attributes"]["updatedAt"]
        self.parent_manga = next((x["id"] for x in rel if x["type"] == "manga"), None)
        self.group = [x["id"] for x in rel if x["type"] == "scanlation_group"]
        self.uploader = next((x["id"] for x in rel if x["type"] == "user"), None)
        self.client = client

    def get_md_network(self, force_443: bool = False):
        return self.client.read_chapter(self, force_443)



from .chapter import Chapter
import time
from requests import exceptions as rex
from pathlib import Path


def dl_page(net, page, path):
    """Helper for dl_chapter to download pages with."""
    try:
        with net.client.session.get(page) as p:
            with open(str(Path(path + "/" + page.rsplit("/", 1)[1])), "wb") as f:
                f.write(p.content)
            success = True if p.status_code <= 400 else False
            cached = True if p.headers["x-cache"] == "HIT" else False
            net.report(page, success, cached, len(p.content), int(p.elapsed.microseconds/1000))
            print(f"Statistics for {page}\nTime: {int(p.elapsed.microseconds/1000)}, length: {len(p.content)}"
                  f"\nSuccess: {success}, was cached on server: {cached}\nHeaders: {p.headers}")
            return True
    except rex.RequestException:
        net.report(page, False, False, 0, 0)
        print(f"Request for {page} failed. This was reported to the MD backend, you should try to download the image"
              f" again to get a new server.")
        return False


def dl_chapter(chapter: Chapter, path, light: bool = False, time_controller: int = 1):
    """Downloads an entire chapter."""
    net = chapter.get_md_network()
    print(f"Got assigned a MD@H node to download: {net.node_url}. Attempting to download {len(net.pages)} pages.")
    if light:
        pages = net.pages_redux
    else:
        pages = net.pages
    for x in pages:
        while True:
            resp = dl_page(net, x, path)
            if resp:
                break
            else:
                net = chapter.get_md_network()
                print(f"Got assigned a new MD@H node to download: {net.node_url}. Resuming downloads...")
        if time_controller:
            time.sleep(time_controller)
    print(f"Successfully downloaded and reported status for {len(net.pages)} pages.")



import time


class NetworkChapter:
    """Represents a link between the MD@H Network and a Chapter."""
    __slots__ = ("valid_thru", "parent_chapter", "node_url", "pages", "pages_redux", "client")

    def __init__(self, chapter, node, client):
        self.valid_thru = int(time.time()) + 900
        self.parent_chapter = chapter
        self.node_url = node
        self.pages = [f"{node}/data/{chapter.hash}/{x}" for x in chapter.pages]
        self.pages_redux = [f"{node}/data-saver/{chapter.hash}/{x}" for x in chapter.pages_redux]
        self.client = client

    def report(self, url, success, cache_header, req_bytes, req_duration):
        return self.client.network_report(url, success, cache_header, req_bytes, req_duration)
